#include <iostream>
#include <cmath>              // mathematical C library
#include <algorithm>             // defining the max() operator
#include <vector>           // STL vector templates
#include <complex> 
using namespace std;

#ifndef PI 
const double PI=3.141592653589793238462643;
#endif 

double N(const double& z) {
    if (z >  6.0) { return 1.0; }; // this guards against overflow 
    if (z < -6.0) { return 0.0; };

    double b1 =  0.31938153; 
    double b2 = -0.356563782; 
    double b3 =  1.781477937;
    double b4 = -1.821255978;
    double b5 =  1.330274429; 
    double p  =  0.2316419; 
    double c2 =  0.3989423; 

    double a=fabs(z); 
    double t = 1.0/(1.0+a*p); 
    double b = c2*exp((-z)*(z/2.0)); 
    double n = ((((b5*t+b4)*t+b3)*t+b2)*t+b1)*t; 
    n = 1.0-b*n; 
    if ( z < 0.0 ) n = 1.0 - n; 
    return n; 
};

double option_price_call_black_scholes(const double& S,       // spot (underlying) price
				       const double& K,       // strike (exercise) price,
				       const double& r,       // interest rate
				       const double& sigma,   // volatility 
				       const double& time) {  // time to maturity 
    double time_sqrt = sqrt(time);
    double d1 = (log(S/K)+r*time)/(sigma*time_sqrt)+0.5*sigma*time_sqrt; 
    double d2 = d1-(sigma*time_sqrt);
    return S*N(d1) - K*exp(-r*time)*N(d2);
};

double option_price_call_european_finite_diff_explicit(const double& S, 
						       const double& K, 
						       const double& r,
						       const double& sigma,
						       const double& time,
						       const int& no_S_steps,
						       const int& no_t_steps) {
    double sigma_sqr = pow(sigma,2);
    int M=no_S_steps; if ((no_S_steps%2)==1) { ++M; } // need no_S_steps to be even:
    double delta_S = 2.0*S/M;
    vector<double> S_values(M+1);
    for (int m=0;m<=M;m++) { S_values[m] = m*delta_S; };
    int N=no_t_steps;
    double delta_t = time/N;
    
    vector<double> a(M);
    vector<double> b(M);
    vector<double> c(M);
    double r1=1.0/(1.0+r*delta_t);
    double r2=delta_t/(1.0+r*delta_t);
    for (unsigned int j=1;j<M;j++){
	a[j] = r2*0.5*j*(-r+sigma_sqr*j);
	b[j] = r1*(1.0-sigma_sqr*j*j*delta_t);
	c[j] = r2*0.5*j*(r+sigma_sqr*j);
    };
    vector<double> f_next(M+1);
    for (int m=0;m<=M;++m) { f_next[m]=max(0.0,S_values[m]-K); };
    vector<double> f(M+1);
    for (int t=N-1;t>=0;--t) {
	f[0]=0;
	for (unsigned m=1;m<M;++m) {
	    f[m]=a[m]*f_next[m-1]+b[m]*f_next[m]+c[m]*f_next[m+1];
	};
	f[M] = 0;
	for (unsigned m=0;m<=M;++m) { f_next[m] = f[m]; };
    };
    return f[M/2];
};


struct heston_parms {double K; double x; double r;  double v; double tau;  double kappa; double theta; 
                     double rho; double sigma; double lambda;  int j;};
 
extern "C"{


double heston_integrand_j(double phi, void *p){ 
    struct heston_parms* parms = (struct heston_parms*)p;
    double K = (parms->K); double x = (parms->x); 
    double v = (parms->v); double r = (parms->r);
    double kappa = (parms->kappa);
    double theta = (parms->theta);
    double rho = (parms->rho);
    double sigma = (parms->sigma);
    double lambda = (parms->lambda);
    double tau = (parms->tau);
    double j = (parms->j);
    double sigma_sqr = pow(sigma,2);
    double uj;    double bj;
    if (j==1){	uj=0.5;  bj=kappa+lambda-rho*sigma; }
    else {      uj=-0.5; bj=kappa+lambda;   };
    std::complex <double> i(0,1);
    double a = kappa*theta;
    std::complex<double> d  = sqrt( pow(rho*sigma*phi*i-bj,2) - sigma_sqr*(2*uj*phi*i-pow(phi,2)) );
    std::complex<double> g  = (bj - rho*sigma*phi*i+d)/(bj-rho*sigma*phi*i-d);
    std::complex<double> C  = r*phi*i*tau+(a/sigma_sqr)*((bj-rho*sigma*phi*i+d)*tau-2.0*log((1.0-g*exp(d*tau))/(1.0-g)));
    std::complex<double> D  = (bj-rho*sigma*phi*i+d)/sigma_sqr * ( (1.0-exp(d*tau))/(1.0-g*exp(d*tau)) );
    std::complex<double> f1 = exp(C+D*v+i*phi*x);
    std::complex<double> F  = exp(-phi*i*log(K))*f1/(i*phi);
	return std::real(F); 
};};


 static double x[150] = {
0.542626,0.370395,0.230993,0.124405,0.050618,0.009607,
0.747706,0.985657,1.256506,1.560281,
1.897016,2.266749,2.669519,3.105372,3.574356,4.076522,4.611926,5.180628,5.782691,6.418183,
7.087175,7.789743,8.525966,9.295927,10.099714,10.937420,11.809140,12.714974,13.655029,14.629413,
15.638241,16.681630,17.759705,18.872593,20.020428,21.203347,22.421494,23.675016,24.964068,26.288808,
27.649400,29.046015,30.478829,31.948023,33.453785,34.996309,36.575795,38.192451,39.846489,41.538130,
43.267601,45.035137,46.840980,48.685380,50.568593,52.490886,54.452532,56.453814,58.495024,60.576461,
62.698436,64.861268,67.065287,69.310833,71.598257,73.927922,76.300199,78.715476,81.174150,83.676631,
86.223344,88.814727,91.451232,94.133325,96.861491,99.636228,102.458051,105.327495,108.245112,
111.211471,114.227165,117.292806,120.409026,123.576484,126.795859,130.067858,133.393212,136.772683,
140.207058,143.697158,147.243835,150.847975,154.510500,158.232369,162.014582,165.858181,169.764254,
173.733934,177.768406,181.868909,186.036738,190.273249,194.579863,198.958070,203.409435,207.935600,
212.538294,217.219337,221.980648,226.824250,231.752283,236.767009,241.870826,247.066279,252.356069,
257.743073,263.230360,268.821204,274.519111,280.327841,286.251434,292.294241,298.460962,304.756684,
311.186934,317.757729,324.475644,331.347894,338.382420,345.588004,352.974402,360.552509,368.334558,
376.334374,384.567687,393.052539,401.809802,410.863871,420.243586,429.983502,440.125674,450.722256,
461.839398,473.563384,486.010824,499.346798,513.820397,529.844189,548.212947,570.989411};

static double w[150] = {
0.188652,0.155814,0.122992,0.090185,0.057392,0.024654,0.221512,0.254395,0.287307,0.320249,0.353228,
0.386245,0.419304,0.452410,0.485566,0.518775,0.552043,0.585372,0.618766,0.652230,0.685767,0.719382,
0.753078,0.786859,0.820731,0.854696,0.888760,0.922927,0.957201,0.991586,1.026088,1.060711,1.095460,
1.130339,1.165353,1.200509,1.235809,1.271261,1.306869,1.342638,1.378575,1.414685,1.450973,1.487446,
1.524110,1.560971,1.598036,1.635311,1.672802,1.710518,1.748464,1.786649,1.825079,1.863763,1.902708,
1.941923,1.981416,2.021196,2.061272,2.101654,2.142349,2.183370,2.224725,2.266426,2.308483,2.350908,
2.393712,2.436908,2.480508,2.524525,2.568974,2.613867,2.659220,2.705048,2.751367,2.798193,2.845543,
2.893437,2.941891,2.990927,3.040563,3.090823,3.141728,3.193301,3.245567,3.298552,3.352283,3.406789,
3.462099,3.518244,3.575258,3.633176,3.692034,3.751871,3.812729,3.874651,3.937683,4.001874,4.067277,
4.133945,4.201939,4.271320,4.342156,4.414519,4.488484,4.564134,4.641557,4.720849,4.802110,4.885451,
4.970991,5.058860,5.149198,5.242157,5.337903,5.436618,5.538501,5.643771,5.752670,5.865463,5.982447,
6.103950,6.230339,6.362026,6.499475,6.643209,6.793823,6.951998,7.118515,7.294276,7.480333,7.677916,
7.888481,8.113766,8.355859,8.617311,8.901266,9.211659,9.553501,9.933297,10.359677,10.844413,11.404090,
12.063035,12.858816,13.853485,15.159514,17.010660,20.015419,26.688256};

inline double heston_Pj(double S, double K, double r, double v, double tau, double sigma, 
                 double kappa, double lambda, double rho, double theta, int j){

	//FILE   *out_abs= fopen("abs.res","w"), *out_wexp= fopen("wexp.res","w");
    double y=log(S);
    struct heston_parms parms = { K, y, r, v, tau, kappa, theta, rho, sigma, lambda, j};
    size_t n=10000;

	int i;

	const int       glsteps = 150; /* <= 150 for double precision */
   
	/*integral Gauss-Laguerre*/
	double sum = 0, phi;
	for (i=glsteps-1;i>=0;--i)
	  {
		phi = x[i];
		sum += w[i] * heston_integrand_j(phi, &parms);
	  }

    return 0.5 + sum/PI;
};



	double heston_call_option_price(const double& S, const double& K, const double& r, const double& v, 
				const double& tau, const double& rho,  const double& kappa, 
				const double& lambda,  const double& theta, const double& sigma){
    double P1 = heston_Pj(S,K,r,v,tau,sigma,kappa,lambda,rho,theta,1);
    double P2 = heston_Pj(S,K,r,v,tau,sigma,kappa,lambda,rho,theta,2);
    double C=S*P1-K*exp(-r*tau)*P2;
    return C;
};

void test_option_price_call_black_scholes(){
    double S = 50; double K = 50; double r = 0.10;
    double sigma = 0.30; double time=0.50;
    cout << " Black Scholes call price = ";
    cout << option_price_call_black_scholes(S, K , r, sigma, time) << endl;
};

 
void black_scholes_examples(){
    cout << "----------------------------------------" << endl;
    cout << "Examples in Black Scholes chapter " << endl; 
    cout << "----------------------------------------" << endl;
    test_option_price_call_black_scholes();
};
 
 
void test_heston(){
    double S=100;
    double K=100;
    double r=0.01;
    double v=0.01;
    double tau=0.5;
    double rho=0;
    double kappa=2;
    double lambda=0.0;
    double theta=0.01;
    double sigma=0.01;
    cout << " Heston call price " << heston_call_option_price( S, K, r, v, tau, rho, kappa, lambda, theta, sigma) << endl;
};

void alternative_formulas_examples(){
    cout << "-----------------------------" << endl;
    cout << "Alternative formulas " << endl;
    cout << "-----------------------------" << endl;
	test_heston();
};
 
void test_explicit_finite_differences(){
    double S = 50.0;
    double K = 50.0;
    double r = 0.1; 
    double sigma = 0.4;
    double time=0.4167;
    int no_S_steps=20;
    int no_t_steps=11;
    cout << " explicit finite differences, european put price = ";
    cout << option_price_call_european_finite_diff_explicit(S,K,r,sigma,time,no_S_steps,no_t_steps) 
	 << endl;
};

void finite_differences_examples(){
    cout << "----------------------------" << endl;
    cout << "Finite Differences examples " << endl;
    cout << "----------------------------" << endl;
    test_explicit_finite_differences();
};

int main() {
    black_scholes_examples();
	alternative_formulas_examples();
	finite_differences_examples();    
};
